"""
Test
~~~~
"""

import unittest
from D365API.Test.TestAccess import TestAccess

def load_tests(loader, tests, pattern):
    """
    Customize how modules or packages load tests during normal test
    runs or test discovery. This `load_tests` function loads all the
    tests in specific order of execution.

    Args:
        loader (TestLoader): Used to create test suites from classes
            and modules.
        tests (TestCase): The tests loaded by default from the module.
        pattern (str): 

    Returns:
        A TestSuite with the individual test cases.
    """

    # Define the test classes
    test_classes = (
        TestAccess,
    )

    # Create the Unit Test Suite
    suite = unittest.TestSuite()

    # Loop through each of the test classes
    for test_class in test_classes:
        # Load the tests from the test class
        tests = loader.loadTestsFromTestCase(test_class)
        # Add the tests to the test suite
        suite.addTests(tests)

    return suite


"""
D365API.TestD365API
~~~~~~~~~~~~~~~~~~~
"""

import json
import os
import unittest

from D365API.D365API import D365API
from D365API.Constant import TEST_FILE


def setUpModule():
    """Set Up Module"""
    pass


def tearDownModule():
    """Tear Down Module"""
    pass


class TestD365APIGeneral(unittest.TestCase):
    """Test."""

    def test_general(self):
        # Get the current directory of the file
        current_directory = os.path.dirname(os.path.abspath(__file__))
        # Get the path of the Test Data file
        test_data_file = os.path.join(current_directory, TEST_FILE)

        # Open the file for reading
        with open(test_data_file, 'r') as f:
            data = json.load(f)

        # Get the hostname from the Test Data
        hostname = data['organizations']['name']

        # Get the user data for success login
        oauth_1_0_user_success = data['systemusers']['oauth_1_0_user_success']

        # Get the read Account failure unique identifier (ID)
        read_account_id = 'af2e07f6-fa17-ec11-b6e6-000d3a9b2012'

        # Create an instance of D365API object and login
        d365api = D365API(hostname=hostname,
                          client_id=oauth_1_0_user_success['client_id'],
                          client_secret=oauth_1_0_user_success['client_secret'],
                          tenant_id=oauth_1_0_user_success['tenant_id'])

        print(f'Label: {d365api.accounts.label}')

        print(f'Account: {d365api.accounts.read(id=read_account_id)}')


if __name__ == '__main__':
    unittest.main()


"""
D365API.Entity
~~~~~~~~~~~~~~
"""

import json
from urllib.parse import urlencode
import requests

from D365API.Constant import D365_API_V

class Entity(object):
    """Entity.
    """

    def __init__(self, access, hostname):
        """Constructor.

        Args:
            access (str): The Microsoft Dynamics 365 access token.
            hostname (str): The Hostname of the environment.
        """

        # Get the access token and set the URL (Uniform Resource Locator)
        self.access_token = access
        self.root_url = f'https://{hostname}.api.crm.dynamics.com/api/data/v{D365_API_V}'

        # Create header
        self.header = {
            'Authorization': 'Bearer ' + self.access_token,
            'Content-Type': 'application/json; charset=utf-8',
            'Accept': 'application/json',
            'OData-Version': '4.0',
            'OData-MaxVersion': '4.0'
        }


    def __getattr__(self, label):
        """Get Attribute Passed In.

        Args:
            label (str): The attribute passed in.

        Returns:
            A instance of the Entity class.
        """
        # Set the name / label
        self.label = label

        # Return the self instance
        return self


    def create(self, payload):
        """Create Entity.

        Args:
            payload (dict): The payload (message body) passed in.

        Returns:
            A string for the unique identifier (ID) of the Entity.
        """

        # Create request URL
        request_url = f'{self.root_url}/{self.label}'

        # Send the request for a response
        r = requests.post(url=request_url,
                          headers=self.header,
                          data=payload)

        # Check the status code
        if r.status_code == 204:
            # Parse the unique identifier (ID) of the entity
            entity_url = r.headers['OData-EntityId']
            begin_id = entity_url.find('(') + 1
            end_id = entity_url.find(')')
            entity_id = entity_url[begin_id:end_id]
            # Return the unique identifier (ID) of the entity
            return entity_id

        # There was an error
        return None


    def read(self, id=None):
        """Read Entity.

        Args:
            id (str): The unique identifier (ID) of the entity.

        Returns:
            A string formatted JSON for the read request
        """

        # Create a read result list to store all the read result
        read_result_list = []

        # Create request URL
        if id is not None:
            # Read a single record
            request_url = f'{self.root_url}/{self.label}({id})'
        else:
            # Read all records
            request_url = f'{self.root_url}/{self.label}'

        # Send the request for a response
        r = requests.get(url=request_url,
                         headers=self.header)

        # Check the failure status code
        if r.status_code != 200:
            return None

        # Check the success status code
        if r.status_code == 200:
            # Parse the read result
            read_result = json.loads(r.text)

            # Add the read result to list
            if 'value' in read_result:
                # Use extend for multiple result
                read_result_list.extend(read_result['value'])
            else:
                # Use append for single result
                read_result_list.append(read_result)

        # Check if there are more result
        while '@odata.nextLink' in read_result:
            # If there are more result
            # Parse the URL for the next set of result
            request_url = read_result['@odata.nextLink']
            # Send the request for a response
            r = requests.get(url=request_url,
                             headers=self.header)

            # Check the status code
            if r.status_code == 200:
                # Parse the read result
                read_result = json.loads(r.text)

                # Add the read result to list
                if 'value' in read_result:
                    # Use extend for multiple result
                    read_result_list.extend(read_result['value'])
                else:
                    # Use append for single result
                    read_result_list.append(read_result)

        # Return all the read result
        return read_result_list


    def update(self, id, payload):
        """Update Entity.

        Args:
            id (str): The unique identifier (ID) of the entity.
            payload (dict): The payload (message body) passed in.

        Returns:
            An integer for the status code of the update request.
        """

        # Create request URL
        request_url = f'{self.root_url}/{self.label}({id})'

        # Update header
        # Don't perform create for this update
        self.header['If-Match'] = '*'

        # Send the request for a response
        r = requests.patch(url=request_url,
                           headers=self.header,
                           data=payload)

        # Check the status code
        if r.status_code == 204:
            # Return the status code
            return r.status_code

        # There was an error
        return None


    def upsert(self, id, payload, option=None):
        """Upsert Entity.
        """

        # TODO: Implementation


    def delete(self, id):
        """Delete Entity.

        Args:
            id (str): The unique identifier (ID) of the entity.

        Returns:
            An integer for the status code of the delete request.
        """

        # Create request URL
        request_url = f'{self.root_url}/{self.label}({id})'

        # Send the request for a response
        r = requests.delete(url=request_url,
                            headers=self.header)

        # Check the status code
        if r.status_code == 204:
            # Return the status code
            return r.status_code

        # There was an error
        return None


    def associate(self, primary_id, collection, secondary, secondary_id, update=False):
        """Associate Entity.

        .. _Associate and Disassociate Table Rows Using The Web API
        https://docs.microsoft.com/en-us/powerapps/developer/data-platform/webapi/associate-disassociate-entities-using-web-api

        Args:
            primary_id (str): The unique identifier (ID) of the primary
                entity.
            collection (str): The collection to associate the entity to.
            secondary (str): The secondary entity name.
            secondary_id (str): The unique identifier (ID) of the
                secondary entity.
            update (bool): Determine whether or not this is an update
                operation to change the reference.
        """

        # Create request URL
        request_url = f'{self.root_url}/{self.label}({primary_id})/{collection}/$ref'

        # Create payload
        if update:
            # Change the reference in a single valued navigation property
            payload = {
                '@odata.id': f'{secondary}({secondary_id})'
            }
        else:
            # Add a reference to a collection value navigation property
            payload = {
                '@odata.id': f'{self.root_url}/{secondary}({secondary_id})'
            }

        # Send the request for a response
        r = requests.post(url=request_url,
                          headers=self.header,
                          data=json.dumps(payload))

        # Check the status code
        if r.status_code == 204:
            # Return the status code
            return r.status_code

        # There was an error
        return None


    def disassociate(self, primary_id, collection, collection_id=None, secondary=None, secondary_id=None):
        """Disassociate Entity.

        .. _Associate and Disassociate Table Rows Using The Web API
        https://docs.microsoft.com/en-us/powerapps/developer/data-platform/webapi/associate-disassociate-entities-using-web-api

        Args:
            primary_id (str): The unique identifier (ID) of the primary
                entity.
            collection (str): The collection to associate the entity to.
            collection_id (str): The unique identifier (ID) of the
                collection entity.
            secondary (str): The secondary entity name.
            secondary_id (str): The unique identifier (ID) of the
                secondary entity.
        """

        # Create request URL
        if secondary and collection_id is None:
            # Disassociate with primary and secondary entity
            request_url = f'{self.root_url}/{self.label}({primary_id})/{collection}/$ref?$id={self.root_url}/{secondary}({secondary_id})'
        elif collection_id and (secondary is None and secondary_id is None):
            # Disassociate with primary and collection entity
            request_url = f'{self.root_url}/{self.label}({primary_id})/{collection}({collection_id})'
        else:
            return None

        # Send the request for a response
        r = requests.delete(url=request_url,
                            headers=self.header)

        # Check the status code
        if r.status_code == 204:
            # Return the status code
            return r.status_code

        # There was an error
        return None


    def query(self, **kwargs):
        """Query Entity.

        .. _Query Data Using The Web API:
        https://docs.microsoft.com/en-us/powerapps/developer/common-data-service/webapi/query-data-web-api

        .. _Web API Query Data Sample:
        https://docs.microsoft.com/en-us/powerapps/developer/common-data-service/webapi/web-api-query-data-sample

        Args:
            kwargs (dict): The keyword arguments for the query.

        Returns:
            A string formatted JSON for the result of the query.
        """

        # Initialize query
        query = ''
        # query = urlencode(kwargs, safe='$,')

        # Build query
        # Loop through `kwargs` parameter(s)
        query = '&'.join(f'${key}={value}' for key, value in kwargs.items())

        # Create request URL
        request_url = f'{self.root_url}/{self.label}?{query}'

        # Send the request for a response
        r = requests.get(url=request_url,
                         headers=self.header)

        # Check the status code
        if r.status_code == 200:
            # Return the response text (message body)
            return r.text

        # There was an error
        return None


"""
D365API.TestAccess
~~~~~~~~~~~~~~~~~~
"""

import json
import os
import unittest
from D365API.Access import Access
from D365API.Constant import TEST_FILE

class TestAccess(unittest.TestCase):
    """Test the Access module."""

    @classmethod
    def setUpClass(cls):
        """Prepare test class.

        Get the Test Data from JSON (JavaScript Object Notation) file.
        """

        # Get the current directory of the file
        current_directory = os.path.dirname(os.path.abspath(__file__))
        # Get the path of the test data file
        test_access_file = os.path.join(current_directory, TEST_FILE)

        # Open the file for reading
        with open(test_access_file, 'r') as f:
            cls.data = json.load(f)

        # Get the hostname
        cls._hostname = cls.data['organizations']['name']


    def test_login_rest_v1_success(self):
        """Test a success of REST (REpresentational State Transfer) login method.

        Get the success username and password (oauth_1_0_user_success)
        from the Test Data file and login. Should result in login method
        returning an access token.
        """

        # Get the user data for success login
        oauth_1_0_user_success = self.data['systemusers']['oauth_1_0_user_success']

        # Create an instance of Access object and login
        access = Access(hostname=self._hostname,
                        client_id=oauth_1_0_user_success['client_id'],
                        client_secret=oauth_1_0_user_success['client_secret'],
                        tenant_id=oauth_1_0_user_success['tenant_id']).login()

        # Test to ensure access is a string
        self.assertEqual(type(access), str)


    def test_login_rest_v1_failure(self):
        """Test a failure of REST (REpresentational State Transfer) login method.

        Get the failure username and password (oauth_1_0_user_failure)
        from the Test Data file and login. Should result in login method
        returning None value.
        """

        # Get the user data for success login
        oauth_1_0_user_failure = self.data['systemusers']['oauth_1_0_user_failure']

        # Create an instance of Access object and login
        access = Access(hostname=self._hostname,
                        client_id=oauth_1_0_user_failure['client_id'],
                        client_secret=oauth_1_0_user_failure['client_secret'],
                        tenant_id=oauth_1_0_user_failure['tenant_id']).login()

        # Test to ensure access is not a string
        self.assertNotEqual(type(access), str)


if __name__ == '__main__':
    unittest.main()


"""
D365API.TestAccount
~~~~~~~~~~~~~~~~~~~
"""

import json
import os
import uuid
import random
import unittest
import requests

from D365API.Access import Access
from D365API.Entity import Entity
from D365API.Constant import TEST_FILE


def setUpModule():
    """Set Up Module"""
    pass


def tearDownModule():
    """Tear Down Module"""
    pass


class TestAccountCreate(unittest.TestCase):
    """Test the Entity module with Create Account."""
    
    @classmethod
    def setUpClass(cls):
        """Prepare test set up class.

        Get the data from JSON (JavaScript Object Notation) file and
        login. Initialize any prerequisite.
        """

        # Get the current directory of the file
        current_directory = os.path.dirname(os.path.abspath(__file__))
        # Get the path of the Test Data file
        cls.test_data_file = os.path.join(current_directory, TEST_FILE)

        # Open the file for reading
        with open(cls.test_data_file, 'r') as f:
            cls.data = json.load(f)

        # Get the hostname from the Test Data
        cls.hostname = cls.data['organizations']['name']

        # Get the user data for success login
        oauth_1_0_user_success = cls.data['systemusers']['oauth_1_0_user_success']

        # Create an instance of Access object and login
        cls.access = Access(hostname=cls.hostname,
                            client_id=oauth_1_0_user_success['client_id'],
                            client_secret=oauth_1_0_user_success['client_secret'],
                            tenant_id=oauth_1_0_user_success['tenant_id']).login()

        # Create an instance of Entity
        cls.entity = Entity(cls.access, cls.hostname)


    def test_create_account_random_failure(self):
        """Test a random failure for create Account.

        Get the hostname from the Test Data and the generated a random
        Account field to make a request for create. Should result in
        response with status code 400 Bad Request.
        """
        
        # Create payload
        payload = {
            # Generate a random Account field
            'random': f'Random-{random.randrange(10000, 99999)}'
        }

        # Make a request to create the Account with random field
        # Create the Account with the newly generated Account field
        create_account_id = self.entity.accounts.create(json.dumps(payload))

        # Test to ensure Account information is None
        self.assertIsNone(create_account_id)


    @unittest.expectedFailure
    def test_create_account_description_failure(self):
        """Test a description failure for create Account.

        Get the hostname from the Test Data and the generated a random
        Account Description to make a request for create. Should result
        in response with status code 400 Bad Request.
        """
        
        # Create payload
        payload = {
            # Generate a random Account description
            'description': f'Description-{random.randrange(10000, 99999)}'
        }

        # Make a request to create the Account with random description
        # Create the Account with the newly generated Account description
        create_account_id = self.entity.accounts.create(json.dumps(payload))

        # Test to ensure Account information is None
        self.assertIsNone(create_account_id)


    def test_create_account_success(self):
        """Test a success for create Account.

        Get the hostname from the Test Data and the generated Account
        Name to make a request for create. Should result in response
        with status code 204 No Content.
        """

        # Create payload
        payload = {
            # Generate a random Account Name
            'name': f'Account-{random.randrange(10000, 99999)}'
        }

        # Make a request to create the Account
        # Get the return unique identifier (ID)
        # The payload need to be serialized to JSON formatted str (json.dumps)
        create_account_id = self.entity.accounts.create(json.dumps(payload))

        # Create the dictionary for Account data
        self.data['accounts'] = {}
        self.data['accounts']['create_account_success'] = {}
        # Create or update the Account ID in the Test Data
        self.data['accounts']['create_account_success']['id'] = create_account_id

        # Write the new Test Data to file
        with open(self.test_data_file, 'w') as f:
            json.dump(self.data, f)

        # Test to ensure Account ID is a string
        self.assertEqual(type(create_account_id), str)


    @classmethod
    def tearDownClass(cls):
        """Prepare test tear down class.

        Clean up Test Data.
        """

        # Check if Account data exist
        if 'accounts' in cls.data:
            # Get the create Account success unique identifier (ID)
            create_account_id = cls.data['accounts']['create_account_success']['id']
            # Make a request to delete the Account
            create_account = cls.entity.accounts.delete(create_account_id)
            # Check if the delete was successful
            if create_account == 204:
                # Delete the Account entry from the Test Data
                del cls.data['accounts']

            # Write the new Test Data to file
            with open(cls.test_data_file, 'w') as f:
                json.dump(cls.data, f)


class TestAccountRead(unittest.TestCase):
    """Test the Entity module with Read Account."""
    
    @classmethod
    def setUpClass(cls):
        """Prepare test set up class.

        Get the data from JSON (JavaScript Object Notation) file and
        login. Initialize any prerequisite.
        """

        # Get the current directory of the file
        current_directory = os.path.dirname(os.path.abspath(__file__))
        # Get the path of the Test Data file
        cls.test_data_file = os.path.join(current_directory, TEST_FILE)

        # Open the file for reading
        with open(cls.test_data_file, 'r') as f:
            cls.data = json.load(f)

        # Get the hostname from the Test Data
        cls.hostname = cls.data['organizations']['name']

        # Get the user data for success login
        oauth_1_0_user_success = cls.data['systemusers']['oauth_1_0_user_success']

        # Create an instance of Access object and login
        cls.access = Access(hostname=cls.hostname,
                            client_id=oauth_1_0_user_success['client_id'],
                            client_secret=oauth_1_0_user_success['client_secret'],
                            tenant_id=oauth_1_0_user_success['tenant_id']).login()

        # Create an instance of Entity
        cls.entity = Entity(cls.access, cls.hostname)

        # Create payload
        payload = {
            # Generate a random Account Name
            'name': f'Account-{random.randrange(10000, 99999)}'
        }

        # Make a request to create the Account
        # Get the return unique identifier (ID)
        # The payload need to be serialized to JSON formatted str (json.dumps)
        cls.account_id = cls.entity.accounts.create(json.dumps(payload))


    def test_read_account_failure(self):
        """Test a failure of read Account.

        Generate a random Universally Unique IDentifier (UUID) to make a
        request for read. Should result in response with None.
        """

        # Generate a random UUID
        read_account_id = uuid.uuid4()

        # Make a request to read the Account with unique identifier (ID)
        account = self.entity.accounts.read(read_account_id)

        # Test to ensure read Account information is None
        self.assertIsNone(account)


    def test_read_account_success(self):
        """Test a success for read Account.

        Get the hostname from the Test Data to make a request for read.
        Should result in response with status code 200 OK and a list
        formatted JSON for the result of the read.
        """

        # Make a request to read the Account
        read_account = self.entity.accounts.read(self.account_id)

        # Test to ensure Account information is a string
        self.assertEqual(type(read_account), list)

    @unittest.expectedFailure
    def test_read_account_count_success(self):
        """Test a success for read count Account.

        Get the hostname from the Test Data to make a request for read.
        Should result in response with a list, the count Account result
        is +1 more than the read Account result from  the
        `RetrieveTotalRecordCount` function due to `setUpClass`
        function.
        """

        # Get the Account count using the `RetrieveTotalRecordCount` function
        request_url = f"{self.entity.accounts.root_url}/RetrieveTotalRecordCount(EntityNames=['account'])"

        # Create header
        header = {
            'Authorization': f'Bearer {self.access}',
            'Content-Type': 'application/json; charset=utf-8',
            'Accept': 'application/json',
            'OData-Version': '4.0',
            'OData-MaxVersion': '4.0'
        }

        # Make a request to count the total Account
        count_account_result = requests.get(url=request_url,
                                            headers=header)

        # Parse the Account count result
        count_account = json.loads(count_account_result.text)['EntityRecordCountCollection']['Values'][0]

        # Make a request to read the Account
        read_account = self.entity.accounts.read()

        # Test to ensure count Account (+1) is the same as read Account
        self.assertEqual((count_account + 1), len(read_account))


    @classmethod
    def tearDownClass(cls):
        """Prepare test tear down class.

        Clean up Test Data.
        """

        # Make a request to delete the Account
        _ = cls.entity.accounts.delete(cls.account_id)


class TestAccountUpdate(unittest.TestCase):
    """Test the Entity module with Update Account."""
    
    @classmethod
    def setUpClass(cls):
        """Prepare test set up class.

        Get the data from JSON (JavaScript Object Notation) file and
        login.
        """

        # Get the current directory of the file
        current_directory = os.path.dirname(os.path.abspath(__file__))
        # Get the path of the Test Data file
        cls.test_data_file = os.path.join(current_directory, TEST_FILE)

        # Open the file for reading
        with open(cls.test_data_file, 'r') as f:
            cls.data = json.load(f)

        # Get the hostname from the Test Data
        cls.hostname = cls.data['organizations']['name']

        # Get the user data for success login
        oauth_1_0_user_success = cls.data['systemusers']['oauth_1_0_user_success']

        # Create an instance of Access object and login
        cls.access = Access(hostname=cls.hostname,
                            client_id=oauth_1_0_user_success['client_id'],
                            client_secret=oauth_1_0_user_success['client_secret'],
                            tenant_id=oauth_1_0_user_success['tenant_id']).login()

        # Create an instance of Entity
        cls.entity = Entity(cls.access, cls.hostname)

        # Create payload
        payload = {
            # Generate a random Account Name
            'name': f'Account-{random.randrange(10000, 99999)}'
        }

        # Make a request to create the Account
        # Get the return unique identifier (ID)
        # The payload need to be serialized to JSON formatted str (json.dumps)
        cls.account_id = cls.entity.accounts.create(json.dumps(payload))


    def test_update_account_failure(self):
        """Test a failure for Account update.

        Generate a random Universally Unique IDentifier (UUID) and an
        incorrect Account Name to make a request for update. Should
        result in response with None.
        """

        # Generate a random UUID
        update_account_id = uuid.uuid4()

        # Create payload
        payload = {
            # Generate a random Account Name
            'name': f'Account-{random.randrange(10000, 99999)}'
        }

        # Make a request to update the Account with the incorrect unique identifier (ID)
        # Update the Account Name with the newly generated Account Name
        update_account = self.entity.accounts.update(update_account_id, json.dumps(payload))

        # Test to ensure update Account information is None
        self.assertIsNone(update_account)


    def test_update_account_success(self):
        """Test a success for Account update.

        Get the hostname from the Test Data and the generated a new
        Account Name to make a request for update. Should result in
        response with status code 204 No Content.
        """

        # Create payload
        payload = {
            # Generate a random Account Name
            'name': f'Account-{random.randrange(10000, 99999)}'
        }

        # Make a request to update the Account with unique identifier (ID)
        # Update the Account Name with the newly generated Account Name
        update_account = self.entity.accounts.update(self.account_id, json.dumps(payload))

        # Test to ensure HTTP status code is 204 No Content
        self.assertEqual(update_account, 204)


    @classmethod
    def tearDownClass(cls):
        """Prepare test tear down class.

        Clean up Test Data.
        """

        # Make a request to delete the Account
        _ = cls.entity.accounts.delete(cls.account_id)


class TestAccountDelete(unittest.TestCase):
    """Test the Entity module with Delete Account."""
    
    @classmethod
    def setUpClass(cls):
        """Prepare test set up class.

        Get the data from JSON (JavaScript Object Notation) file and
        login.
        """

        # Get the current directory of the file
        current_directory = os.path.dirname(os.path.abspath(__file__))
        # Get the path of the Test Data file
        cls.test_data_file = os.path.join(current_directory, TEST_FILE)

        # Open the file for reading
        with open(cls.test_data_file, 'r') as f:
            cls.data = json.load(f)

        # Get the hostname from the Test Data
        cls.hostname = cls.data['organizations']['name']

        # Get the user data for success login
        oauth_1_0_user_success = cls.data['systemusers']['oauth_1_0_user_success']

        # Create an instance of Access object and login
        cls.access = Access(hostname=cls.hostname,
                            client_id=oauth_1_0_user_success['client_id'],
                            client_secret=oauth_1_0_user_success['client_secret'],
                            tenant_id=oauth_1_0_user_success['tenant_id']).login()

        # Create an instance of Entity
        cls.entity = Entity(cls.access, cls.hostname)

        # Create payload
        payload = {
            # Generate a random Account Name
            'name': f'Account-{random.randrange(10000, 99999)}'
        }

        # Make a request to create the Account
        # Get the return unique identifier (ID)
        # The payload need to be serialized to JSON formatted str (json.dumps)
        cls.account_id = cls.entity.accounts.create(json.dumps(payload))


    def test_delete_account_failure(self):
        """Test a failure for Account delete.

        Generate a random Universally Unique IDentifier (UUID) to make a
        request for delete. Should result in response with None.
        """

        # Generate a random UUID
        delete_account_id = uuid.uuid4()

        # Make a request to delete the Account with unique identifier (ID)
        # Delete the Account
        delete_account = self.entity.accounts.delete(delete_account_id)

        # Test to ensure delete Account information is None
        self.assertIsNone(delete_account)


    def test_delete_account_success(self):
        """Test a success for Account delete.

        Get the hostname and the unique identifier (ID) from the Test
        Data to make a request for delete. Should result in response
        with status code 204 No Content.
        """

        # Make a request to delete the Account with unique identifier (ID)
        delete_account = self.entity.accounts.delete(self.account_id)

        # Test to ensure HTTP status code is 204 No Content
        self.assertEqual(delete_account, 204)


    @classmethod
    def tearDownClass(cls):
        """Prepare test tear down class.

        Clean up Test Data.
        """

        # Make a request to delete the Account
        _ = cls.entity.accounts.delete(cls.account_id)


class TestAccountAssociate(unittest.TestCase):
    """Test the Entity module with Associate Account."""

    @classmethod
    def setUpClass(cls):
        """Prepare test set up class.

        Get the data from JSON (JavaScript Object Notation) file and
        login.
        """

        # Get the current directory of the file
        current_directory = os.path.dirname(os.path.abspath(__file__))
        # Get the path of the Test Data file
        cls.test_data_file = os.path.join(current_directory, TEST_FILE)

        # Open the file for reading
        with open(cls.test_data_file, 'r') as f:
            cls.data = json.load(f)

        # Get the hostname from the Test Data
        cls.hostname = cls.data['organizations']['name']

        # Get the user data for success login
        oauth_1_0_user_success = cls.data['systemusers']['oauth_1_0_user_success']

        # Create an instance of Access object and login
        cls.access = Access(hostname=cls.hostname,
                            client_id=oauth_1_0_user_success['client_id'],
                            client_secret=oauth_1_0_user_success['client_secret'],
                            tenant_id=oauth_1_0_user_success['tenant_id']).login()

        # Create an instance of Entity
        cls.entity = Entity(cls.access, cls.hostname)

        # Generate a random number
        random_number = random.randrange(10000, 99999)

        # Create Account using payload
        payload = {
            'name': f'Account-{random_number}'
        }

        # Make a request to create the Account
        # Get the return unique identifier (ID)
        # The payload need to be serialized to JSON formatted str (json.dumps)
        cls.account_id = cls.entity.accounts.create(json.dumps(payload))

        # Create Opportunity using payload
        payload = {
            'name': f'Opportunity-{random_number}'
        }

        # Make a request to create the Opportunity
        cls.opportunity_id = cls.entity.opportunities.create(json.dumps(payload))


    def test_associate_account_opportunity_failure(self):
        """Test a failure for associate Account to Opportunity.

        Generate random Universally Unique IDentifier (UUID) for Account
        and Opportunity to make a request to associate the two entity.
        Should result in a response with None.
        """

        # Generate a random UUID for Account
        random_account_id = uuid.uuid4()
        # Generate a random UUID for Opportunity
        random_opportunity_id = uuid.uuid4()

        # Make a request to associate Account and Opportunity
        associate_account = self.entity.accounts.associate(primary_id=random_account_id,
                                                           collection='opportunity_customer_accounts',
                                                           secondary='opportunities',
                                                           secondary_id=random_opportunity_id)

        # Test to ensure associate Account information is None
        self.assertIsNone(associate_account)


    def test_associate_account_opportunity_success(self):
        """Test a success for associate Account to Opportunity.

        Get the unique identifier (ID) for Account and Opportunity to
        make a request to associate the two entity. Should result in a
        response with status code 204 No Content.
        """

        # Make a request to associate Account and Opportunity
        associate_account = self.entity.accounts.associate(primary_id=self.account_id,
                                                           collection='opportunity_customer_accounts',
                                                           secondary='opportunities',
                                                           secondary_id=self.opportunity_id)

        # Test to ensure HTTP status code is 204 No Content
        self.assertEqual(associate_account, 204)


    @classmethod
    def tearDownClass(cls):
        """Prepare test tear down class.

        Clean up Test Data.
        """

        # Make a request to disassociate Account and Opportunity
        _ = cls.entity.accounts.disassociate(primary_id=cls.account_id,
                                             collection='opportunity_customer_accounts',
                                             secondary='opportunities',
                                             secondary_id=cls.opportunity_id)

        # Make a request to delete the Account
        _ = cls.entity.accounts.delete(cls.account_id)
        # Make a request to delete the Opportunity
        _ = cls.entity.opportunities.delete(cls.opportunity_id)


class TestAccountDisassociate(unittest.TestCase):
    """Test the Entity module with Disassociate Account."""

    @classmethod
    def setUpClass(cls):
        """Prepare test set up class.

        Get the data from JSON (JavaScript Object Notation) file and
        login.
        """

        # Get the current directory of the file
        current_directory = os.path.dirname(os.path.abspath(__file__))
        # Get the path of the Test Data file
        cls.test_data_file = os.path.join(current_directory, TEST_FILE)

        # Open the file for reading
        with open(cls.test_data_file, 'r') as f:
            cls.data = json.load(f)

        # Get the hostname from the Test Data
        cls.hostname = cls.data['organizations']['name']

        # Get the user data for success login
        oauth_1_0_user_success = cls.data['systemusers']['oauth_1_0_user_success']

        # Create an instance of Access object and login
        cls.access = Access(hostname=cls.hostname,
                            client_id=oauth_1_0_user_success['client_id'],
                            client_secret=oauth_1_0_user_success['client_secret'],
                            tenant_id=oauth_1_0_user_success['tenant_id']).login()

        # Create an instance of Entity
        cls.entity = Entity(cls.access, cls.hostname)

        # Generate a random number
        random_number = random.randrange(10000, 99999)

        # Create Account using payload
        payload = {
            'name': f'Account-{random_number}'
        }

        # Make a request to create the Account
        # Get the return unique identifier (ID)
        # The payload need to be serialized to JSON formatted str (json.dumps)
        cls.account_id = cls.entity.accounts.create(json.dumps(payload))

        # Create Opportunity using payload
        payload = {
            'name': f'Opportunity-{random_number}'
        }

        # Make a request to create the Opportunity
        cls.opportunity_id = cls.entity.opportunities.create(json.dumps(payload))

        # Make a request to associate Account and Opportunity
        cls.entity.accounts.associate(primary_id=cls.account_id,
                                      collection='opportunity_customer_accounts',
                                      secondary='opportunities',
                                      secondary_id=cls.opportunity_id)


    def test_disassociate_account_opportunity_failure(self):
        """Test a failure for disassociate Account to Opportunity.

        Generate random Universally Unique IDentifier (UUID) for Account
        and Opportunity to make a request to disassociate the two
        entity. Should result in a response with None.
        """

        # Generate a random UUID for Account
        random_account_id = uuid.uuid4()
        # Generate a random UUID for Opportunity
        random_opportunity_id = uuid.uuid4()

        # Make a request to disassociate Account and Opportunity
        disassociate_account = self.entity.accounts.disassociate(primary_id=random_account_id,
                                                                 collection='opportunity_customer_accounts',
                                                                 secondary='opportunities',
                                                                 secondary_id=random_opportunity_id)

        # Test to ensure disassociate Account information is None
        self.assertIsNone(disassociate_account)


    def test_disassociate_account_opportunity_success(self):
        """Test a success for disassociate Account to Opportunity.

        Get the unique identifier (ID) for Account and Opportunity to
        make a request to disassociate the two entity. Should result in
        a response with status code 204 No Content.
        """

        # Make a request to disassociate Account and Opportunity
        disassociate_account = self.entity.accounts.disassociate(primary_id=self.account_id,
                                                                 collection='opportunity_customer_accounts',
                                                                 secondary='opportunities',
                                                                 secondary_id=self.opportunity_id)

        # Test to ensure HTTP status code is 204 No Content
        self.assertEqual(disassociate_account, 204)


    @classmethod
    def tearDownClass(cls):
        """Prepare test tear down class.

        Clean up Test Data.
        """

        # Make a request to disassociate Account and Opportunity
        _ = cls.entity.accounts.disassociate(primary_id=cls.account_id,
                                             collection='opportunity_customer_accounts',
                                             secondary='opportunities',
                                             secondary_id=cls.opportunity_id)

        # Make a request to delete the Account
        _ = cls.entity.accounts.delete(cls.account_id)
        # Make a request to delete the Opportunity
        _ = cls.entity.opportunities.delete(cls.opportunity_id)


class TestAccountQuery(unittest.TestCase):
    """Test the Entity module with Query Account."""

    @classmethod
    def setUpClass(cls):
        """Prepare test set up class.

        Get the data from JSON (JavaScript Object Notation) file and
        login.
        """

        # Get the current directory of the file
        current_directory = os.path.dirname(os.path.abspath(__file__))
        # Get the path of the Test Data file
        cls.test_data_file = os.path.join(current_directory, TEST_FILE)

        # Open the file for reading
        with open(cls.test_data_file, 'r') as f:
            cls.data = json.load(f)

        # Get the hostname from the Test Data
        cls.hostname = cls.data['organizations']['name']

        # Get the user data for success login
        oauth_1_0_user_success = cls.data['systemusers']['oauth_1_0_user_success']

        # Create an instance of Access object and login
        cls.access = Access(hostname=cls.hostname,
                            client_id=oauth_1_0_user_success['client_id'],
                            client_secret=oauth_1_0_user_success['client_secret'],
                            tenant_id=oauth_1_0_user_success['tenant_id']).login()

        # Create an instance of Entity
        cls.entity = Entity(cls.access, cls.hostname)


    def test_query_select_account_success(self):
        """Test a success for Account query `select`.

        Get the hostname from the Test Data to make a request for query.
        Test for the `select` system query option. Should result in
        response with status code 200 OK and a string formatted JSON for
        the result of the query.
        """

        # Define query property
        query = {
            'select': 'accountid,name'
        }

        # Make a request to query the Account
        query_account = self.entity.accounts.query(**query)

        # Test to ensure Account information is a string
        self.assertEqual(type(query_account), str)


    def test_query_top_account_success(self):
        """Test a success for Account query `top`.

        Get the hostname from the Test Data to make a request for query.
        Test for the `top` system query option. Should result in
        response with status code 200 OK and a string formatted JSON for
        the result of the query.
        """

        # Set the `top` system query option count
        top_count = 3
        
        # Define query property
        query = {
            'top': top_count
        }

        # Make a request to query the Account
        query_account = self.entity.accounts.query(**query)

        # Count the number of result
        query_account_count = len(json.loads(query_account)['value'])

        # Test to ensure the Account query count is same as `top` count
        self.assertLessEqual(query_account_count, top_count)


    def test_query_select_top_account_success(self):
        """Test a success for Account query `select` and `top`.

        Get the hostname from the Test Data to make a request for query.
        Test for the `select` and `top` system query option. Should
        result in response with status code 200 OK and a string
        formatted JSON for the result of the query.
        """

        # Set the `top` system query option count
        top_count = 3

        # Define query property
        query = {
            'select': 'accountid,name',
            'top': top_count
        }

        # Make a request to query the Account
        query_account = self.entity.accounts.query(**query)

        # Test to ensure Account information is a string
        self.assertEqual(type(query_account), str)


    @classmethod
    def tearDownClass(cls):
        """Prepare test tear down class.

        Clean up Test Data.
        """

        # No need clean up
        pass


def suite():
    """Test Suite"""

    # Create the Unit Test Suite
    suite = unittest.TestSuite()

    # Add the Unit Test
    suite.addTest(TestAccountCreate())
    suite.addTest(TestAccountRead())
    suite.addTest(TestAccountUpdate())
    suite.addTest(TestAccountDelete())
    suite.addTest(TestAccountAssociate())
    suite.addTest(TestAccountQuery())

    # Return the Test Suite
    return suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())


"""
D365API
~~~~~~~
"""

from D365API.Access import Access
from D365API.Entity import Entity
from D365API.Constant import D365_API_V

class D365API(Entity):
    """D365API Object

    Main object for the Microsoft D365 (Dynamics 365) API (Application
    Programing Interface).
    """

    def __init__(self, hostname, client_id, client_secret, tenant_id):
        """Constructor.

        Args:
            hostname (str): The Hostname of the environment.
            client_id (str): The Application (client) ID of the application.
            client_secret (str): The client secret of the application.
            tenant_id (str): The Directory (tenant) ID of the application.
        """

        # Create an instance of Access object and login
        access = Access(hostname=hostname,
                        client_id=client_id,
                        client_secret=client_secret,
                        tenant_id=tenant_id).login()

        # Set the root URL (Uniform Resource Locator)
        self.root_url = f'https://{hostname}.api.crm.dynamics.com/api/data/v{D365_API_V}'

        # Create header
        self.header = {
            'Authorization': 'Bearer ' + access,
            'Content-Type': 'application/json; charset=utf-8',
            'Accept': 'application/json',
            'OData-Version': '4.0',
            'OData-MaxVersion': '4.0'
        }



from setuptools import setup, find_packages

# Read the contents of the README.md file
from pathlib import Path
current_directory = Path(__file__).parent
long_description = (current_directory/'README.md').read_text()

setup(
    name = 'd365fw',
    version = '0.8.0',
    packages = find_packages(),

    # Dependency
    install_requires = [
        'requests',
    ],

    # Metadata
    author = 'Yan Kuang',
    author_email = 'YTKme@Outlook.com',
    description = 'Microsoft Dynamics 365 Application Programming Interface.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license = 'GNU GENERAL PUBLIC LICENSE',
    keywords = 'dynamics d365 fw api microsoft'
)


"""
D365API.Access
~~~~~~~~~~~~~~
"""

import requests

class Access(object):
    """Access.
    """

    def __init__(self, hostname, client_id, client_secret, tenant_id):
        """Constructor.

        Args:
            hostname (str): The Hostname of the environment.
            client_id (str): The Application (client) ID of the application.
            client_secret (str): The client secret of the application.
            tenant_id (str): The Directory (tenant) ID of the application.
        """

        self.client_id = client_id
        self.client_secret = client_secret
        self.version = 1
        self.oauth_1_0_url = f'https://login.microsoftonline.com/{tenant_id}/oauth2/token'
        self.resource = f'https://{hostname}.api.crm.dynamics.com/'
        self.oath_2_0_url = f'https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token'
        self.scope = f'https://{hostname}.api.crm.dynamics.com/.default'
    

    def login(self):
        """Login to Microsoft Dynamics 365.

        Returns:
            A string for the Microsoft Dynamics 365 access (bearer)
            token depending whether if request is OAuth 1.0 or OAuth
            2.0.
        """

        if self.version == 1:
            return self.login_oauth_1_0()
        elif self.version == 2:
            return self.login_oauth_2_0()
        else:
            return None


    def login_oauth_1_0(self):
        """Login via REST (REpresentational State Transfer) Version 1.

        Returns:
            A string for the Microsoft Dynamics 365 access (bearer) token if
            call was successful, or None otherwise.
        """

        # Create header
        header = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }

        # Create payload
        payload = {
            'grant_type': 'client_credentials',
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'resource': self.resource
        }

        # Send the request
        r = requests.post(url=self.oauth_1_0_url,
                          headers=header,
                          data=payload)

        # Check the status code
        if r.status_code == 200:
            # Parse the access token
            access_token = r.json()['access_token']
            return access_token

        # There was an error
        return None


    def login_oauth_2_0(self):
        pass
